/*
 * traceroute host  - trace the route ip packets follow going to "host".
 *
 * Attempt to trace the route an ip packet would follow to some
 * internet host.  We find out intermediate hops by launching probe
 * packets with a small ttl (time to live) then listening for an
 * icmp "time exceeded" reply from a gateway.  We start our probes
 * with a ttl of one and increase by one until we get an icmp "port
 * unreachable" (which means we got to "host") or hit a max (which
 * defaults to 30 hops & can be changed with the -m flag).  Three
 * probes (change with -q flag) are sent at each ttl setting and a
 * line is printed showing the ttl, address of the gateway and
 * round trip time of each probe.  If the probe answers come from
 * different gateways, the address of each responding system will
 * be printed.  If there is no response within a 3 sec. timeout
 * interval (changed with the -w flag), a "*" is printed for that
 * probe.
 * If the -l flag is used the ttl from the icmp reply will
 * be printed, otherwise it will be printed only if it has an
 * unexpected value.
 *
 * Probe packets are UDP format.  We don't want the destination
 * host to process them so the destination port is set to an
 * unlikely value (if some clod on the destination is using that
 * value, it can be changed with the -p flag).
 *
 * A sample use might be:
 *
 *     [yak 71]% traceroute nis.nsf.net.
 *     traceroute to nis.nsf.net (35.1.1.48), 30 hops max, 56 byte packet
 *      1  helios.ee.lbl.gov (128.3.112.1)  19 ms  19 ms  0 ms
 *      2  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  39 ms  19 ms
 *      3  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  39 ms  19 ms
 *      4  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  39 ms  40 ms  39 ms
 *      5  ccn-nerif22.Berkeley.EDU (128.32.168.22)  39 ms  39 ms  39 ms
 *      6  128.32.197.4 (128.32.197.4)  40 ms  59 ms  59 ms
 *      7  131.119.2.5 (131.119.2.5)  59 ms  59 ms  59 ms
 *      8  129.140.70.13 (129.140.70.13)  99 ms  99 ms  80 ms
 *      9  129.140.71.6 (129.140.71.6)  139 ms  239 ms  319 ms
 *     10  129.140.81.7 (129.140.81.7)  220 ms  199 ms  199 ms
 *     11  nic.merit.edu (35.1.1.48)  239 ms  239 ms  239 ms
 *
 * Note that lines 2 & 3 are the same.  This is due to a buggy
 * kernel on the 2nd hop system -- lbl-csam.arpa -- that forwards
 * packets with a zero ttl.
 *
 * A more interesting example is:
 *
 *     [yak 72]% traceroute allspice.lcs.mit.edu.
 *     traceroute to allspice.lcs.mit.edu (18.26.0.115), 30 hops max
 *      1  helios.ee.lbl.gov (128.3.112.1)  0 ms  0 ms  0 ms
 *      2  lilac-dmc.Berkeley.EDU (128.32.216.1)  19 ms  19 ms  19 ms
 *      3  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  19 ms  19 ms
 *      4  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  19 ms  39 ms  39 ms
 *      5  ccn-nerif22.Berkeley.EDU (128.32.168.22)  20 ms  39 ms  39 ms
 *      6  128.32.197.4 (128.32.197.4)  59 ms  119 ms  39 ms
 *      7  131.119.2.5 (131.119.2.5)  59 ms  59 ms  39 ms
 *      8  129.140.70.13 (129.140.70.13)  80 ms  79 ms  99 ms
 *      9  129.140.71.6 (129.140.71.6)  139 ms  139 ms  159 ms
 *     10  129.140.81.7 (129.140.81.7)  199 ms  180 ms  300 ms
 *     11  129.140.72.17 (129.140.72.17)  300 ms  239 ms  239 ms
 *     12  * * *
 *     13  128.121.54.72 (128.121.54.72)  259 ms  499 ms  279 ms
 *     14  * * *
 *     15  * * *
 *     16  * * *
 *     17  * * *
 *     18  ALLSPICE.LCS.MIT.EDU (18.26.0.115)  339 ms  279 ms  279 ms
 *
 * (I start to see why I'm having so much trouble with mail to
 * MIT.)  Note that the gateways 12, 14, 15, 16 & 17 hops away
 * either don't send ICMP "time exceeded" messages or send them
 * with a ttl too small to reach us.  14 - 17 are running the
 * MIT C Gateway code that doesn't send "time exceeded"s.  God
 * only knows what's going on with 12.
 *
 * The silent gateway 12 in the above may be the result of a bug in
 * the 4.[23]BSD network code (and its derivatives):  4.x (x <= 3)
 * sends an unreachable message using whatever ttl remains in the
 * original datagram.  Since, for gateways, the remaining ttl is
 * zero, the icmp "time exceeded" is guaranteed to not make it back
 * to us.  The behavior of this bug is slightly more interesting
 * when it appears on the destination system:
 *
 *      1  helios.ee.lbl.gov (128.3.112.1)  0 ms  0 ms  0 ms
 *      2  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  19 ms  39 ms
 *      3  lilac-dmc.Berkeley.EDU (128.32.216.1)  19 ms  39 ms  19 ms
 *      4  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  39 ms  40 ms  19 ms
 *      5  ccn-nerif35.Berkeley.EDU (128.32.168.35)  39 ms  39 ms  39 ms
 *      6  csgw.Berkeley.EDU (128.32.133.254)  39 ms  59 ms  39 ms
 *      7  * * *
 *      8  * * *
 *      9  * * *
 *     10  * * *
 *     11  * * *
 *     12  * * *
 *     13  rip.Berkeley.EDU (128.32.131.22)  59 ms !  39 ms !  39 ms !
 *
 * Notice that there are 12 "gateways" (13 is the final
 * destination) and exactly the last half of them are "missing".
 * What's really happening is that rip (a Sun-3 running Sun OS3.5)
 * is using the ttl from our arriving datagram as the ttl in its
 * icmp reply.  So, the reply will time out on the return path
 * (with no notice sent to anyone since icmp's aren't sent for
 * icmp's) until we probe with a ttl that's at least twice the path
 * length.  I.e., rip is really only 7 hops away.  A reply that
 * returns with a ttl of 1 is a clue this problem exists.
 * Traceroute prints a "!" after the time if the ttl is <= 1.
 * Since vendors ship a lot of obsolete (DEC's Ultrix, Sun 3.x) or
 * non-standard (HPUX) software, expect to see this problem
 * frequently and/or take care picking the target host of your
 * probes.
 *
 * Other possible annotations after the time are !H, !N, !P (got a host,
 * network or protocol unreachable, respectively), !S or !F (source
 * route failed or fragmentation needed -- neither of these should
 * ever occur and the associated gateway is busted if you see one).  If
 * almost all the probes result in some kind of unreachable, traceroute
 * will give up and exit.
 *
 * Still other annotations are (ttl="n") and (ttl="n"!), where "n" is the
 * ttl from the icmp reply packet we received.  The first form will be
 * printed for every packet if you use the -l option.  The second form
 * will be printed (with or without -l) if this ttl is an unexpected value.
 * We expect that the return path from the n'th hop will contain n hops,
 * otherwise we know the reply packet is coming back via a different path than
 * it went out on.  Unfortunately, not everyone uses the same starting ttl
 * value for icmp messages.  The common ones used by routers are 29
 * (Proteon 8.1 and lower software), 59 (Proteon 8.2), 255 (cisco, BSD
 * since 4.3 tahoe).  30 and 60 are also often used by hosts, and probably
 * by some routers, because they were the BSD TCP ttl values.  This makes
 * some "off by one" return paths hard to detect, you might try removing
 * OLD_BSD_TCP and NEW_BSD_TCP from the case statement if this annoys you.
 * If you are using the lsrr (-g) code with the -l code you will see many
 * bogus "!".  
 *
 * With the -l option you will see the TTL value of the ICMP msg that
 * came back to you, as in:
 *
 * nettlerash> traceroute -l rip.berkeley.edu
 * traceroute to rip.berkeley.edu (128.32.131.22), 30 hops max, 40 byte packets
 *  1  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  0 ms (ttl=59)  10 ms (ttl=59)
 *  2  ccn-nerif35.Berkeley.EDU (128.32.168.35)  10 ms (ttl=58)  10 ms (ttl=58)
 *  3  csgw.Berkeley.EDU (128.32.133.254)  10 ms (ttl=253)  20 ms (ttl=253)
 *  4  rip.Berkeley.EDU (128.32.131.22)  30 ms (ttl=252)  30 ms (ttl=252)
 *
 * This shows that from nettlerash the route goes through two proteons (ttl=
 * 59, then 58), followed by a router using max_ttl to rip which also uses
 * max_ttl.  (-l and printing ttl stuff added by cliff@berkeley.edu.)
 *
 * Notes
 * -----
 * This program must be run by root or be setuid.  (I suggest that
 * you *don't* make it setuid -- casual use could result in a lot
 * of unnecessary traffic on our poor, congested nets.)
 *
 * This program requires a kernel mod that does not appear in any
 * system available from Berkeley:  A raw ip socket using proto
 * IPPROTO_RAW must interpret the data sent as an ip datagram (as
 * opposed to data to be wrapped in a ip datagram).  See the README
 * file that came with the source to this program for a description
 * of the mods I made to /sys/netinet/raw_ip.c.  Your mileage may
 * vary.  But, again, ANY 4.x (x < 4) BSD KERNEL WILL HAVE TO BE
 * MODIFIED TO RUN THIS PROGRAM.
 *
 * The udp port usage may appear bizarre (well, ok, it is bizarre).
 * The problem is that an icmp message only contains 8 bytes of
 * data from the original datagram.  8 bytes is the size of a udp
 * header so, if we want to associate replies with the original
 * datagram, the necessary information must be encoded into the
 * udp header (the ip id could be used but there's no way to
 * interlock with the kernel's assignment of ip id's and, anyway,
 * it would have taken a lot more kernel hacking to allow this
 * code to set the ip id).  So, to allow two or more users to
 * use traceroute simultaneously, we use this task's pid as the
 * source port (the high bit is set to move the port number out
 * of the "likely" range).  To keep track of which probe is being
 * replied to (so times and/or hop counts don't get confused by a
 * reply that was delayed in transit), we increment the destination
 * port number before each probe.
 *
 * Don't use this as a coding example.  I was trying to find a
 * routing problem and this code sort-of popped out after 48 hours
 * without sleep.  I was amazed it ever compiled, much less ran.
 *
 * I stole the idea for this program from Steve Deering.  Since
 * the first release, I've learned that had I attended the right
 * IETF working group meetings, I also could have stolen it from Guy
 * Almes or Matt Mathis.  I don't know (or care) who came up with
 * the idea first.  I envy the originators' perspicacity and I'm
 * glad they didn't keep the idea a secret.
 *
 * Tim Seaver, Ken Adelman and C. Philip Wood provided bug fixes and/or
 * enhancements to the original distribution.
 *
 * I've hacked up a round-trip-route version of this that works by
 * sending a loose-source-routed udp datagram through the destination
 * back to yourself.  Unfortunately, SO many gateways botch source
 * routing, the thing is almost worthless.  Maybe one day...
 *
 *  -- Van Jacobson (van@helios.ee.lbl.gov)
 *     Tue Dec 20 03:50:13 PST 1988
 *
 * Copyright (c) 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char Version[] = "@(#)traceroute.c	e07@nikhef.nl (Eric Wassenaar) 930211";
#endif lint

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netdb.h>

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	64	/* maximum length of hostname */
#endif
#ifndef MAXDNAME
#define MAXDNAME	256	/* maximum length of domain name */
#endif

#define	MAX_START_TTL	255	/* Cisco and BSD4.3 */
#define	OLD_PROTEON_TTL	29	/* Proteon 8.1 and lower */
#define	NEW_PROTEON_TTL	59	/* Proteon 8.2 and higher */
#define	OLD_BSD_TCP	30
#define	NEW_BSD_TCP	60

#ifndef FD_SET
#define NFDBITS         (8*sizeof(fd_set))
#define FD_SETSIZE      NFDBITS
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      bzero((char *)(p), sizeof(*(p)))
#endif

#define sameword(a,b)	(strcasecmp(a,b) == 0)

#define NOT_DOTTED_QUAD	((u_long)-1)

#define TRUE	1
#define FALSE	0

#ifdef lint
#define EXTERN
#else
#define EXTERN extern
#endif

EXTERN int errno;

extern void exit();
extern char *index();
extern char *rindex();
extern char *strcpy();
extern char *malloc();
extern char *inet_ntoa();
extern u_long inet_addr();

u_long getaddr();
char *inetname();
char *pr_type();

/*
 * format of a (udp) probe packet.
 */
struct probe
{
	struct ip ip;
	struct udphdr udp;
	u_char seq;		/* sequence number of this packet */
	u_char ttl;		/* ttl packet left with */
	struct timeval tv;	/* time packet left */
};

#define MAXPKT	65535		/* max ip packet size */
#define MAXDATA	(MAXPKT - sizeof(struct probe))

struct probe *opacket;		/* last output (udp) packet */
u_char ipacket[4096];		/* last inbound (icmp) packet */

int sndsock;			/* send (udp) socket file descriptor */
int rcvsock;			/* receive (icmp) socket file descriptor */

struct sockaddr whereto;	/* who to try to reach */
struct sockaddr wherefrom;	/* who we got response from */

u_short ident;			/* packet ident */
u_short port = 32768+666;	/* start udp dest port # for probe packets */
int datalen = 0;		/* how much data in raw ip packet */
int udplen = 0;			/* how much data in udp packet */
int tos = 0;			/* type of service */

char hnamebuf[MAXDNAME+1];
char *hostname;			/* remote host to query about */
char *source = NULL;		/* specific source for multi-homed sender */
char *program;			/* name of this program */

int nprobes = 3;		/* number of probe packets to send */
int max_ttl = OLD_BSD_TCP;	/* max number of hops */

int options = 0;		/* socket options */
int waittime = 3;		/* time to wait for response (in seconds) */
int verbose = FALSE;		/* print additional information */
int numeric = FALSE;		/* print addresses numerically */
int ttlflag = FALSE;		/* print ttl values */

#ifndef MAX_IPOPTLEN
#define MAX_IPOPTLEN 40
#endif

#define MAXIPOPT 9		/* MAX_IPOPTLEN-IPOPT_MINOFF / sizeof(u_long) */
#define MAXLSRR MAXIPOPT-1	/* leave room for destination address */
u_long lsrraddr[MAXLSRR];	/* loose source route addresses */
int nlsrr = 0;			/* count of loose source route addresses */

char Usage[] =
"\
Usage: %s [-v] [-l] [-m maxhops] [-q nqueries] [-g gateway] host\n\
Other: [-n] [-r] [-w waittime] [-p port] [-t tos] [-s src_addr] [data_size]\n\
";


main(argc, argv)
int argc;
char *argv[];
{
	struct sockaddr_in *from = (struct sockaddr_in *)&wherefrom;
	struct sockaddr_in *to   = (struct sockaddr_in *)&whereto;
	int on = 1;
	int probe;
	int ttl;
	int seq = 0;
	u_long addr;
	struct hostent *hp;
	struct protoent *proto;
	register char *opt;

	program = rindex(argv[0], '/');
	if (program++ == NULL)
		program = argv[0];

#if defined(SYSV_SETVBUF)
	setvbuf(stdout, NULL, _IOLBF, BUFSIZ);
#else
	setlinebuf(stdout);
#endif

/*
 * Check command line options.
 */
	while (argc > 2 && argv[1][0] == '-')
	{
	    for (opt = &argv[1][1]; *opt; opt++)
	    {
		switch (*opt)
		{
		    case 'd':
			options |= SO_DEBUG;
			break;

		    case 'g':
			if (nlsrr >= MAXLSRR)
				fatal("no more than %d gateways", MAXLSRR);
			argc--, argv++;
			lsrraddr[nlsrr] = getaddr(argv[1]);
			nlsrr++;
			break;

		    case 'l':
			ttlflag = TRUE;
			break;

		    case 'm':
			argc--, argv++;
			max_ttl = atoi(argv[1]);
			if (max_ttl <= 1)
				fatal("maxhops must be > 1");
			break;

		    case 'n':
			numeric = TRUE;
			break;

		    case 'p':
			argc--, argv++;
			port = atoi(argv[1]);
			if (port < 1)
				fatal("port must be > 0");
			break;

		    case 'q':
			argc--, argv++;
			nprobes = atoi(argv[1]);
			if (nprobes < 1)
				fatal("nprobes must be > 0");
			break;

		    case 'r':
			options |= SO_DONTROUTE;
			break;

		    case 's':
			argc--, argv++;
			source = argv[1];
			break;

		    case 't':
			argc--, argv++;
			tos = atoi(argv[1]);
			if (tos < 0 || tos > 255)
				fatal("tos must be 0 to 255");
			break;

		    case 'v':
			verbose = TRUE;
			break;

		    case 'w':
			argc--, argv++;
			waittime = atoi(argv[1]);
			if (waittime <= 1)
				fatal("wait must be > 1 sec");
			break;

		    default:
			(void) fprintf(stderr, Usage, program);
			exit(1);
		}
	    }

	    argc--, argv++;
	}

	if (argc < 2 || argv[1][0] == '-')
	{
		(void) fprintf(stderr, Usage, program);
		exit(1);
	}

/*
 * Fetch remote host address to probe.
 */
	hostname = argv[1];
	addr = inet_addr(hostname);
	if (addr == NOT_DOTTED_QUAD)
	{
		hp = gethostbyname(hostname);
		if (hp == NULL)
			fatal("unknown host %s", hostname);
		hostname = strcpy(hnamebuf, hp->h_name);

		to->sin_family = hp->h_addrtype;
		to->sin_port = 0;
		bcopy(hp->h_addr, (char *)&to->sin_addr, hp->h_length);
	}
	else
	{
		hp = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
		if (hp != NULL)
			hostname = hp->h_name;
		hostname = strcpy(hnamebuf, hostname);

		to->sin_family = AF_INET;
		to->sin_port = 0;
		to->sin_addr.s_addr = addr;
	}

/*
 * Check for optional extra data size of output packet.
 */
	if (argc > 2)
	{
		datalen = atoi(argv[2]);
		if (datalen < 0 || datalen > MAXDATA)
			fatal("packet size must be 0 to %d", MAXDATA);
	}

/*
 * Allocate output packet of proper size.
 */
	datalen += sizeof(struct probe);
	opacket = (struct probe *)malloc((unsigned)datalen);
	if (opacket == NULL)
		fatal("out of memory");

	(void) bzero((char *)opacket, datalen);
	opacket->ip.ip_dst = to->sin_addr;

	udplen = datalen - sizeof(struct ip);
	ident = (getpid() & 0xFFFF) | 0x8000;

/*
 * Allocate input socket to receive replies.
 */
	proto = getprotobyname("icmp");
	if (proto == NULL)
		fatal("unknown protocol: icmp");

	rcvsock = socket(AF_INET, SOCK_RAW, proto->p_proto);
	if (rcvsock < 0)
		fperror("icmp socket");

	if (options & SO_DEBUG)
		(void) setsockopt(rcvsock, SOL_SOCKET, SO_DEBUG, (char *)&on, sizeof(on));

	if (options & SO_DONTROUTE)
		(void) setsockopt(rcvsock, SOL_SOCKET, SO_DONTROUTE, (char *)&on, sizeof(on));

/*
 * Allocate output socket to send probe packets.
 */
#ifdef SYSV_UDPPROTO
	sndsock = socket(AF_INET, SOCK_RAW, IPPROTO_UDP);
#else
	sndsock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
#endif SYSV_UDPPROTO
	if (sndsock < 0)
		fperror("raw socket");

#ifdef SO_SNDBUF
	if (setsockopt(sndsock, SOL_SOCKET, SO_SNDBUF, (char *)&datalen, sizeof(datalen)) < 0)
		fperror("SO_SNDBUF");
#endif SO_SNDBUF

#ifdef IP_HDRINCL
	if (setsockopt(sndsock, IPPROTO_IP, IP_HDRINCL, (char *)&on, sizeof(on)) < 0)
		fperror("IP_HDRINCL");
#endif IP_HDRINCL

	if (options & SO_DEBUG)
		(void) setsockopt(sndsock, SOL_SOCKET, SO_DEBUG, (char *)&on, sizeof(on));

	if (options & SO_DONTROUTE)
		(void) setsockopt(sndsock, SOL_SOCKET, SO_DONTROUTE, (char *)&on, sizeof(on));

#ifdef IP_OPTIONS
/*
 * Set ip options if loose source routing addresses were specified.
 */
	if (nlsrr > 0)
	{
		register int i = nlsrr + 1;
		char optval[MAX_IPOPTLEN];
		char *optptr;

		bzero(optval, sizeof(optval));
		optval[IPOPT_OPTVAL] = IPOPT_LSRR;
		optval[IPOPT_OLEN]   = IPOPT_MINOFF-1 + i*sizeof(u_long);
		optval[IPOPT_OFFSET] = IPOPT_MINOFF;

		optptr = &optval[IPOPT_MINOFF-1];
		for (i = 0; i < nlsrr; i++)
		{
			bcopy((char *)&lsrraddr[i], optptr, sizeof(u_long));
			optptr += sizeof(u_long);
		}
		bcopy((char *)&to->sin_addr, optptr, sizeof(u_long));

		proto = getprotobyname("ip");
		if (proto == NULL)
			fatal("unknown protocol: ip");

		if (setsockopt(sndsock, proto->p_proto, IP_OPTIONS, optval, MAX_IPOPTLEN) < 0)
			fperror("IP_OPTIONS");
	}
#else
	if (nlsrr > 0)
		error("gateway source routing not supported");
#endif IP_OPTIONS

/*
 * Fill in source address in output packet, if specified.
 */
	if (source)
	{
		addr = inet_addr(source);
		if (addr == NOT_DOTTED_QUAD)
			fatal("illegal address %s", source);

		from->sin_family = AF_INET;
		from->sin_port = 0;
		from->sin_addr.s_addr = addr;

		opacket->ip.ip_src = from->sin_addr;
#ifndef IP_HDRINCL
		if (bind(sndsock, &wherefrom, sizeof(wherefrom)) < 0)
			fperror("bind");
#endif IP_HDRINCL
	}

/*
 * Start of main loop.
 */
	printf("traceroute to %s (%s)", hostname, inet_ntoa(to->sin_addr));
	if (source)
		printf(" from %s", source);
	printf(" %d hops max, %d byte packets\n", max_ttl, datalen);

	for (ttl = 1; ttl <= max_ttl; ttl++)
	{
		u_long lastaddr = 0;
		int timeouts = 0;
		int got_there = 0;
		int unreachable = 0;

		printf("%2d ", ttl);

		for (probe = 1; probe <= nprobes; probe++)
		{
			int cc;
			int status;
			struct timeval sendtime;
			struct timeval recvtime;
			struct ip *ip;

			(void) fflush(stdout);

			(void) gettimeofday(&sendtime, (struct timezone *)NULL);

			seq++;
			(void) send_probe(sndsock, &whereto, seq, ttl);
retry:
			cc = wait_for_reply(rcvsock, &wherefrom, waittime);
			if (cc <= 0)
			{
				printf(" *");
				timeouts++;
				continue;	/* next probe */
			}

			(void) gettimeofday(&recvtime, (struct timezone *)NULL);

			status = packet_ok(ipacket, cc, from, seq);
			if (status < 0)
				goto retry;	/* not our reply */

			if (from->sin_addr.s_addr != lastaddr)
			{
				print_from(ipacket, cc, from);
				lastaddr = from->sin_addr.s_addr;
			}

			printf("  %d ms", tvdelta(&recvtime, &sendtime));

			ip = (struct ip *)ipacket;
			switch(ip->ip_ttl + (ttl - 1))
			{
			    case MAX_START_TTL:
			    case OLD_PROTEON_TTL:
			    case NEW_PROTEON_TTL:
			    case OLD_BSD_TCP:
			    case NEW_BSD_TCP:
				if (ttlflag)
					printf(" (ttl=%d)", ip->ip_ttl);
				break;

			    default:
				if (ttlflag || (nlsrr == 0))
					printf(" (ttl=%d!)", ip->ip_ttl);
				break;
			}

			switch(status)
			{
			    case ICMP_UNREACH_PORT:
				if (ip->ip_ttl <= 1)
					printf(" !");
				got_there++;
				break;

			    case ICMP_UNREACH_PROTOCOL:
				printf(" !P");
				got_there++;
				break;

			    case ICMP_UNREACH_NET:
				printf(" !N");
				unreachable++;
				break;

			    case ICMP_UNREACH_HOST:
				printf(" !H");
				unreachable++;
				break;

			    case ICMP_UNREACH_NEEDFRAG:
				printf(" !F");
				unreachable++;
				break;

			    case ICMP_UNREACH_SRCFAIL:
				printf(" !S");
				unreachable++;
				break;
			}

		} /* for probe <= nprobes */

		printf("\n");

		if (got_there)
			break;
		if (unreachable && (unreachable + timeouts >= nprobes))
			break;

	} /* for ttl <= max_ttl */

	exit(0);
	/*NOTREACHED*/
}


/*VARARGS1*/
error(fmt, a, b, c, d)
char *fmt;
{
	(void) fprintf(stderr, "%s: ", program);
	(void) fprintf(stderr, fmt, a, b, c, d);
	(void) fprintf(stderr, "\n");
}



/*VARARGS1*/
fatal(fmt, a, b, c, d)
char *fmt;
{
	error(fmt, a, b, c, d);
	exit(1);
}


fperror(fmt)
char *fmt;
{
	(void) fprintf(stderr, "%s: ", program);
	perror(fmt);
	exit(1);
}


/*
** GETADDR -- Fetch internet address of host
** -----------------------------------------
**
**	Returns:
**		Internet address of given host.
**		Aborts if address could not be determined.
*/

u_long
getaddr(host)
char *host;
{
	u_long addr;
	struct hostent *hp;

	addr = inet_addr(host);
	if (addr != NOT_DOTTED_QUAD)
		return(addr);

	hp = gethostbyname(host);
	if (hp == NULL)
		fatal("unknown host %s", host);

	bcopy(hp->h_addr, (char *)&addr, hp->h_length);
	return(addr);
}


/*
** SEND_PROBE -- Send out probe packet
** -----------------------------------
**
**	Inputs:
**		Global variable ident   contains our identification.
**		Global variable port    contains the base port offset.
**		Global variable tos     contains the type of service.
**		Global variable opacket points to the output packet.
**		Global variable datalen contains size of raw ip packet.
**		Global variable udplen  contains size of udp packet.
**
**	Returns:
**		Number of bytes sent.
**		-1 in case of error.
*/

send_probe(sock, toaddr, seq, ttl)
int sock;				/* socket descriptor */
struct sockaddr *toaddr;		/* internet address to send to */
int seq;				/* sequence no. determines port */
int ttl;				/* current ttl value */
{
	struct probe  *pkt = opacket;
	struct ip     *ip  = &pkt->ip;
	struct udphdr *udp = &pkt->udp;
	int addrlen;
	int n;

/*
 * Fill in ip header.
 */
	/* ip->ip_v   = (u_char)  IPVERSION;	/* version */
	/* ip->ip_hl  = (u_char)  0;		/* header length */
	   ip->ip_tos = (u_char)  tos;		/* type of service */
	   ip->ip_len = (short)   datalen;	/* total size */
	/* ip->ip_id  = (u_short) 0;		/* identification */
	   ip->ip_off = (short)   0;		/* fragment offset */
	   ip->ip_ttl = (u_char)  ttl;		/* time-to-live */
	   ip->ip_p   = (u_char)  IPPROTO_UDP;	/* protocol */
	   ip->ip_sum = (u_short) 0;		/* checksum */
	/* ip->ip_src = INADDR_ANY;		/* source address */
	/* ip->ip_dst = toaddr->sin_addr;	/* destination address */

/*
 * Fill in udp header.
 */
	udp->uh_sport = htons(ident);		/* source port */
	udp->uh_dport = htons(port+seq);	/* destination port */
	udp->uh_ulen  = htons((u_short)udplen);	/* udp size */
	udp->uh_sum   = 0;			/* checksum */

/*
 * Fill in remainder of output packet.
 */
	pkt->seq = (u_char)seq;
	pkt->ttl = (u_char)ttl;
	(void) gettimeofday(&pkt->tv, (struct timezone *)NULL);

/*
 * Send out the probe packet.
 */
	addrlen = sizeof(*toaddr);
#ifdef SYSV_UDPPROTO
	if (setsockopt(sock, IPPROTO_IP, IP_TTL, (char *)&ttl, sizeof(ttl)) < 0)
		fperror("IP_TTL");

	n = sendto(sock, (char *)udp, udplen, 0, toaddr, addrlen);
	if (n > 0)
		n += sizeof(struct ip);
#else
	n = sendto(sock, (char *)opacket, datalen, 0, toaddr, addrlen);
#endif SYSV_UDPPROTO
	if (n < 0 || n != datalen)
	{
		if (n < 0)
			perror("\nsendto");
		else
			error("\n%d chars instead of %d to %s",
				n, datalen, hostname);
		return(-1);
	}
	return(n);
}


/*
** WAIT_FOR_REPLY -- Wait for arrival of input packet
** --------------------------------------------------
**
**	Outputs:
**		Input packet is stored in global buffer ipacket.
**
**	Returns:
**		Number of bytes received.
**		-1 in case of error or timeout.
*/

wait_for_reply(sock, fromaddr, secs)
int sock;				/* socket descriptor */
struct sockaddr *fromaddr;		/* buffer to store internet address */
int secs;				/* timeout value in secs */
{
	fd_set fds;
	struct timeval wait;
	int n;
	int addrlen;

	wait.tv_sec = secs;
	wait.tv_usec = 0;

	/* FD_ZERO(&fds); */
	bzero((char *)&fds, sizeof(fds));
	FD_SET(sock, &fds);

/*
 * Wait for arrival of input packet, or timeout.
 */
	n = select(FD_SETSIZE, &fds, (fd_set *)0, (fd_set *)0, &wait);
	if (n <= 0)
	{
		if (n == 0)
			errno = ETIMEDOUT;
		return(-1);
	}

/*
 * Fake an error if nothing was actually read.
 */
	addrlen = sizeof(*fromaddr);
	n = recvfrom(sock, (char *)ipacket, sizeof(ipacket), 0, fromaddr, &addrlen);
	if (n == 0)
		errno = ECONNRESET;
	return(n);
}


/*
** PACKET_OK -- Check validity of received input packet
** ----------------------------------------------------
**
**	Inputs:
**		Global variable ident contains our identification.
**		Global variable port  contains the base port offset.
**
**	Returns:
**		ICMP_TIMXCEED if ttl dropped to zero (intermediate host).
**		Subcode of ICMP_UNREACH if destination unreachable.
**		Got to the final destination if ICMP_UNREACH_PORT.
**		-1 if not one of the above.
*/

packet_ok(buf, cc, from, seq)
u_char *buf;				/* address of input packet */
int cc;					/* total size of input packet */
struct sockaddr_in *from;		/* remote internet address */
int seq;				/* sequence no. determines port */
{
	struct ip *ip;
	struct icmp *icmp;
	struct udphdr *udp;
	int iphdrlen;
	u_char type;			/* type of icmp input packet */
	u_char code;			/* subcode of type */

/*
 * The input packet contains the ip header in front of the icmp packet.
 * Make sure the packet contains the icmp header after the ip header.
 */
	ip = (struct ip *)buf;
	iphdrlen = ip->ip_hl << 2;

	if (cc < iphdrlen + ICMP_MINLEN)
	{
		if (verbose)
			printf("\npacket too short (%d bytes) from %s\n",
				cc, inet_ntoa(from->sin_addr));
		return(-1);
	}

	cc -= iphdrlen;
	icmp = (struct icmp *)(buf + iphdrlen);

/*
 * Check the type of the icmp packet.
 * We should get ICMP_TIMXCEED if we got to an intermediate host.
 * We get ICMP_UNREACH if the desired host is currently unreachable,
 * or if we got to the final destination itself.
 * Make sure this is really our response packet.
 * The original ip packet is stored in the icmp message after the
 * icmp header, with the original udp header after the ip header.
 */
	type = icmp->icmp_type;
	code = icmp->icmp_code;

	if ((type == ICMP_TIMXCEED && code == ICMP_TIMXCEED_INTRANS) ||
	     type == ICMP_UNREACH)
	{
		ip = &icmp->icmp_ip;	/* original ip packet */
		iphdrlen = ip->ip_hl << 2;
		udp = (struct udphdr *)((u_char *)ip + iphdrlen);

		if (cc >= (ICMP_MINLEN + iphdrlen + 2*sizeof(u_short)) &&
		    ip->ip_p == IPPROTO_UDP &&
		    udp->uh_sport == htons(ident) &&
		    udp->uh_dport == htons(port+seq))
			return(type == ICMP_TIMXCEED ? type : code);
	}

/*
 * Useless package. Print contents.
 */
	if (verbose)
	{
		register int i;
		u_char *datap = (u_char *)&icmp->icmp_data[0];

		ip = (struct ip *)buf;
		printf("\n%d bytes from %s to %s, ",
			cc, inet_ntoa(from->sin_addr), inet_ntoa(ip->ip_dst));
		printf("icmp type %d (%s) code %d\n",
			type, pr_type(type), code);
		for (i = ICMP_MINLEN; i < cc ; i++)
			printf(" %2.2x", *datap++);
		printf("\n");
	}

	return(-1);
}


/*
** PR_TYPE -- Return type of icmp input packet
** -------------------------------------------
**
**	Returns:
**		Pointer to name of packet type.
*/

char *
pr_type(type)
u_char type;				/* type value of icmp packet */
{
	static char *ttab[] =
	{
	/*  0 */	"echo reply",
	/*  1 */	"ICMP 1",
	/*  2 */	"ICMP 2",
	/*  3 */	"destination unreachable",
	/*  4 */	"source quench",
	/*  5 */	"redirect",
	/*  6 */	"ICMP 6",
	/*  7 */	"ICMP 7",
	/*  8 */	"echo request",
	/*  9 */	"ICMP 9",
	/* 10 */	"ICMP 10",
	/* 11 */	"time exceeded",
	/* 12 */	"parameter problem",
	/* 13 */	"timestamp request",
	/* 14 */	"timestamp reply",
	/* 15 */	"info request",
	/* 16 */	"info reply",
	/* 17 */	"mask request",
	/* 18 */	"mask reply"
	};

	if (type > 18)
		return("unknown packet type");

	return(ttab[type]);
}

 
/*
** PRINT_FROM -- Print the address from which we got an input packet
** -----------------------------------------------------------------
**
**	Returns:
**		None.
*/

print_from(buf, cc, from)
u_char *buf;				/* address of input packet */
int cc;					/* number of bytes received */
struct sockaddr_in *from;		/* remote internet address */
{
	struct ip *ip;
	int iphdrlen;

	ip = (struct ip *)buf;
	iphdrlen = ip->ip_hl << 2;
	cc -= iphdrlen;

	if (numeric)
		printf(" %s", inet_ntoa(from->sin_addr));
	else
		printf(" %s (%s)", inetname(from->sin_addr), inet_ntoa(from->sin_addr));

	if (verbose)
		printf (" %d bytes to %s", cc, inet_ntoa(ip->ip_dst));
}


#ifdef notyet

/*
** CHECKSUM -- Checksum routine for Internet Protocol family headers
** -----------------------------------------------------------------
**
**	Returns:
**		Value of checksum over given buffer.
*/

u_short
checksum(buf, len)
u_short *buf;				/* address of packet */
int len;				/* length of packet */
{
	register int nleft = len;	/* remaining short words */
	register u_short *w = buf;	/* address of next short word */
	register int sum = 0;		/* 32 bit accumulator */
	u_short answer = 0;

/*
 * Our algorithm is simple, using a 32 bit accumulator (sum),
 * we add sequential 16 bit words to it, and at the end, fold back
 * all the carry bits from the top 16 bits into the lower 16 bits.
 */
	while (nleft > 1)
	{
		sum += *w++;
		nleft -= 2;
	}

/*
 * Mop up an odd byte, if necessary.
 */
	if (nleft == 1)
	{
		*(u_char *)(&answer) = *(u_char *)w;
		sum += answer;
	}

/*
 * Add back carry outs from top 16 bits to low 16 bits.
 */
	sum = (sum >> 16) + (sum & 0xFFFF);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */

	return(answer);
}

#endif notyet


/*
** TVDELTA -- Subtract two timeval structs
** ---------------------------------------
**
**	Returns:
**		Value of time difference in milliseconds.
*/

tvdelta(t2, t1)
struct timeval *t2;			/* latest timeval */
struct timeval *t1;			/* oldest timeval */
{
	if ((t2->tv_usec -= t1->tv_usec) < 0)
	{
		t2->tv_sec--;
		t2->tv_usec += 1000000;
	}
	t2->tv_sec -= t1->tv_sec;

	return(t2->tv_sec*1000 + (t2->tv_usec + 500)/1000);
}


/*
** INETNAME -- Construct internet address representation
** -----------------------------------------------------
**
**	Returns:
**		Pointer to string containing address representation.
*/

char *
inetname(inaddr)
struct in_addr inaddr;			/* internet address to convert */
{
	static char myhostname[MAXHOSTNAMELEN+1];
	static char *domain = NULL;	/* our domain name, if available */
	static int gotdomain = FALSE;	/* set if we queried our domain */
	static char buf[MAXDNAME+1];	/* buffer to store representation */
	struct hostent *hp = NULL;
	u_long addr;
	register char *cp;

/*
 * Try to figure out our own domain name.
 */
	if (!numeric && !gotdomain)
	{
		if (gethostname(myhostname, MAXHOSTNAMELEN) == 0)
		{
			domain = index(myhostname, '.');
			if (domain != NULL)
				domain++;
		}

		gotdomain = TRUE;
	}

/*
 * Look up the hostname corresponding with the given address.
 */
	addr = inaddr.s_addr;
	if (!numeric && (addr != INADDR_ANY))
	{
		hp = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
		if (hp != NULL)
		{
			cp = index(hp->h_name, '.');
			if ((cp != NULL) && domain && sameword(cp+1, domain))
				*cp = '\0';
		}
	}

/*
 * Construct the proper address representation.
 */
	if (hp != NULL)
		(void) strcpy(buf, hp->h_name);
	else
		(void) sprintf(buf, "%s", inet_ntoa(inaddr));

	return(buf);
}
